//--------------------------------------------------------------------------------------------------------------------//
//                                                MEMORY FILE ACCESS WITH C LIB                                       //
// 13/02/02, Mythos                                                                                                   //
//--------------------------------------------------------------------------------------------------------------------//

//----------------------------------------------- INCLUDES -----------------------------------------------------------//

#include  "stdafx.h"
#include  "MemFileC.h"
#ifdef    _DEBUG
#include  "MemFileC.inl"
#endif

#include  "Global/Typedefs.h"
#include  "Memory/Allocators/Allocator.h"

//--------------------------------------------------------------------------------------------------------------------//
//                                                CONSTRUCTORS / DESTRUCTOR                                           //
//--------------------------------------------------------------------------------------------------------------------//

//----------------------------------------------- Init ---------------------------------------------------------------//
// init all members
// 14/02/02, Mythos
// in :
// out:
//--------------------------------------------------------------------------------------------------------------------//

void Mythos::CMemFileC::Init()
  {
  m_pAlloc   = NULL;
  m_pu8File  = m_pu8CurrPos = NULL;
  m_u32Size  = m_u32Used    = 0;

  m_boError  = m_boEOF = false;
  m_boClosed = true;
  m_u32Flags = 0;

  m_strName  = "";
  }

//----------------------------------------------- DeleteBuffer -------------------------------------------------------//
// delete buffer if any
// 14/02/02, Mythos
// in :
// out:
//--------------------------------------------------------------------------------------------------------------------//

void Mythos::CMemFileC::DeleteBuffer()
  {
  Memory::Free(m_pu8File,m_pAlloc);

  m_pu8File = m_pu8CurrPos = NULL;
  m_u32Size = m_u32Used    = 0;
  }

//--------------------------------------------------------------------------------------------------------------------//
//                                                OPERATIONS                                                          //
//--------------------------------------------------------------------------------------------------------------------//

//----------------------------------------------- Open ---------------------------------------------------------------//
// open memory file
// 14/02/02, Mythos
// in : name,mode,allocator
// out: OK?
// rem: in read/append mode Open tries to load the file from disk
//--------------------------------------------------------------------------------------------------------------------//

bool Mythos::CMemFileC::Open(const char* pszName,const FILE_MODE Mode,CAllocator* pAlloc)
  {
  if(m_pu8File) Close();                                    // previous file
  DeleteBuffer();                                           // reset
  Init();

  if(pszName && (Mode&_READ_ || Mode&_APPEND_))
    {                                                       // try auto-load
    Load(pszName,NULL,pAlloc);
    }

  m_pAlloc   = pAlloc;
  m_boClosed = false;
  m_u32Flags = Mode;
  m_strName  = pszName;

  return true;
  }

//----------------------------------------------- Read ---------------------------------------------------------------//
// read u32Size bytes to pBuffer
// 14/02/02, Mythos
// in : buffer to fill,nb bytes
// out: nb bytes read
//--------------------------------------------------------------------------------------------------------------------//

u32 Mythos::CMemFileC::Read(void* pvBuffer,const u32 u32Size)
  {
  if(!m_pu8File || m_boClosed)
    {
    m_boError = true;
    return 0;
    }

  u32 u32Read = u32Size;
  if(m_pu8CurrPos+u32Read > m_pu8File+m_u32Used)
    {                                                       // beyond end of file
    m_boEOF = true;
    u32Read = m_pu8File+m_u32Used-m_pu8CurrPos;
    }

  Memory::MemCpy(pvBuffer,m_pu8CurrPos,u32Read);
  m_pu8CurrPos += u32Read;
  return u32Read;
  }

//----------------------------------------------- Write --------------------------------------------------------------//
// write u32Size bytes from pBuffer
// 15/02/02, Mythos
// in : src buffer,nb bytes
// out: nb bytes written
//--------------------------------------------------------------------------------------------------------------------//

u32 Mythos::CMemFileC::Write(const void* pvBuffer,const u32 u32Size)
  {
  if(!pvBuffer || !u32Size) return 0;

  if(!m_pu8File || (m_pu8CurrPos+u32Size > m_pu8File+m_u32Size))
    {                                                       // current buffer is too small
    if(!Expand(m_pu8CurrPos-m_pu8File+u32Size)) return 0;
    }

  Memory::MemCpy(m_pu8CurrPos,pvBuffer,u32Size);
  m_pu8CurrPos += u32Size;
  if(u32(m_pu8CurrPos-m_pu8File) > m_u32Used) m_u32Used = m_pu8CurrPos-m_pu8File;
  return u32Size;
  }

//----------------------------------------------- Expand -------------------------------------------------------------//
// expand memory file
// 15/02/02, Mythos
// in : new minimum size
// out: OK?
//--------------------------------------------------------------------------------------------------------------------//

bool Mythos::CMemFileC::Expand(const u32 u32Size)
  {
  if(u32Size <= m_u32Size) return true;                     // already big enough
  u32 u32NewSize = 1024;                                    // 1Ko mini
  while(u32NewSize < u32Size) u32NewSize <<= 1;             // power of 2

  void* pvNewBuffer = Memory::Malloc(u32NewSize,m_pAlloc);
  if(!pvNewBuffer) return false;

  if(m_pu8File) Memory::MemCpy(pvNewBuffer,m_pu8File,m_u32Used);

  u32 u32Used = m_u32Used;
  u32 u32Pos  = m_pu8CurrPos-m_pu8File;
  DeleteBuffer();

  m_pu8File    = (u8*)pvNewBuffer;
  m_u32Size    = u32NewSize;
  m_u32Used    = u32Used;
  m_pu8CurrPos = m_pu8File+u32Pos;
  return true;
  }

//----------------------------------------------- Close --------------------------------------------------------------//
// close file
// 14/02/02, Mythos
// in :
// out: OK?
// rem: auto-saves the file if the flag has been passed to Open
//--------------------------------------------------------------------------------------------------------------------//

bool Mythos::CMemFileC::Close()
  {
  if(m_boClosed) return m_boError = true;

  bool boSaved = true;
  if(m_u32Flags & _AUTOSAVE_) boSaved = Save(m_strName.c_str(),m_pu8File,m_u32Size);

  m_boClosed = true;
  return boSaved;
  }

//----------------------------------------------- Seek ---------------------------------------------------------------//
// modify current position pointer
// 14/02/02, Mythos
// in : offset,origin
// out: 0=OK
//--------------------------------------------------------------------------------------------------------------------//

u32 Mythos::CMemFileC::Seek(const u32 u32Pos,const u32 u32Orig)
  {
  if(!m_pu8File || m_boClosed) return m_boError = true;
  u32 u32Offset = u32(-1);

  if(u32Orig == SEEK_SET) u32Offset = u32Pos;
  if(u32Orig == SEEK_CUR) u32Offset = m_pu8CurrPos-m_pu8File+u32Pos;
  if(u32Orig == SEEK_END) u32Offset = (u32)m_pu8File+m_u32Used+u32Pos;

  if(u32Offset > (u32)m_pu8File+m_u32Used) return m_boError = true;
  m_pu8CurrPos = m_pu8File+u32Offset;
  return 0;
  }

//----------------------------------------------- Tell ---------------------------------------------------------------//
// return current position offset
// 14/02/02, Mythos
// in :
// out: -1=error
//--------------------------------------------------------------------------------------------------------------------//

u32 Mythos::CMemFileC::Tell()
  {
  if(!m_pu8File || m_boClosed)
    {
    m_boError = true;
    return -1;
    }
  return m_pu8CurrPos-m_pu8File;
  }

//----------------------------------------------- GetString ----------------------------------------------------------//
// read string
// 15/02/02, Mythos
// in : dest buffer,its size
// out: OK?
//--------------------------------------------------------------------------------------------------------------------//

bool Mythos::CMemFileC::GetString(char* pszString,const u32 u32Len)
  {
  if(!pszString) return false;

  u32  u32Read = 0;
  u32  u32Left = GetSize()-Tell();                          // bytes left to read
  bool boChr13 = false;

  while((u32Read < u32Len-1) && u32Left)
    {
    u8 u8Chr = *m_pu8CurrPos++;
    u32Left--;
    pszString[u32Read++] = u8Chr;

    if(boChr13 && (u8Chr == 10))
      {                                                     // carriage return
      pszString[u32Read-2] = 0;
      return true;
      }
    boChr13 = (u8Chr == 13);
    }

  pszString[u32Read] = 0;
  if(u32Read >= u32Len-1) return true;                      // buffer is full
  
  m_boEOF = true;
  return(u32Read > 0);
  }

//----------------------------------------------- PutString ----------------------------------------------------------//
// write string
// 15/02/02, Mythos
// in : src buffer
// out: OK?
//--------------------------------------------------------------------------------------------------------------------//

bool Mythos::CMemFileC::PutString(const char* pszString)
  {
  if(!pszString) return false;
  u32 u32Len = Mythos::String::StrLen(pszString);

  u32  u32Written = Write(pszString,u32Len);
  bool boRes      = PutChar(13);
  boRes          &= PutChar(10);

  return((u32Written == u32Len) && boRes);
  }

//----------------------------------------------- Load ---------------------------------------------------------------//
// load file from disk
// 15/02/02, Mythos
// in : name,returned size,allocator
// out: memory buffer
//--------------------------------------------------------------------------------------------------------------------//

void* Mythos::CMemFileC::Load(const char* pszName,u32* pu32Size,CAllocator* pAlloc)
  {
  File Loader;
  u32  u32Size;
  void* pvBuff = Loader.Load(pszName,&u32Size,pAlloc);
  if(!pvBuff) return NULL;

  DeleteBuffer();
  Init();

  m_pu8CurrPos = m_pu8File = (u8*)pvBuff;
  m_u32Size    = m_u32Used = u32Size;
  if(pu32Size)   *pu32Size = u32Size;
  return pvBuff;
  }

//----------------------------------------------- Save ---------------------------------------------------------------//
// save to disk
// 15/02/02, Mythos
// in : name,buffer,size
// out: OK?
// rem: buffer/size can be NULL/0 or it should correspond to a subpart of the file
//--------------------------------------------------------------------------------------------------------------------//

bool Mythos::CMemFileC::Save(const char* pszName,const void* pvBuffer,const u32 u32Size)
  {
  if(!pszName) return false;

  u8* pu8Buffer = (u8*)pvBuffer;
  if(!pu8Buffer) pu8Buffer = m_pu8File;                     // pBuffer==NULL => start at the beginning
  if( pu8Buffer < m_pu8File) pu8Buffer = m_pu8File;         // avoid to start before the buffer
  if(!pu8Buffer) return false;

  u32 u32Write = u32Size;
  if(!pvBuffer && !u32Size) u32Write = m_u32Used;           // NULL/0 => full buffer
  if(!u32Write) return false;

  if(pu8Buffer+u32Write > m_pu8File+m_u32Used)              // avoid to go beyond EOF
    u32Write = m_u32Used-(pu8Buffer-m_pu8File);

  File Saver;
  return Saver.Save(pszName,pu8Buffer,u32Write);
  }
